Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.
Require Import Relation_Definitions.
Require Export ImpAlg.Lattices.
Require Export ImpAlg.Ens.

(*************************************)
(*    Definitions of ⊥ and ⊤         *)
(*************************************)

Generalizable Variables X ord join meet arrow L meet_set join_set.



Section BotTop.
  Context `{BL:BoundedLattice}.


  Lemma meet_bot_l: ∀ a, ⊥ ⋀ a = ⊥.
  Proof.
    intros.
    apply (@ord_antisym _ _ (⊥ ⋀ a) ⊥ ).
    apply meet_lb_l.
    apply bot_min.
  Qed.

  Lemma meet_bot_r: ∀ a, a ⋀ ⊥ = ⊥.
  Proof.
    intros.
    apply ord_antisym.
    apply meet_lb_r.
    apply bot_min.
  Qed.

  Lemma join_bot_l: ∀ a, ⊥ ⋁ a = a.
  Proof.
    intros.
    apply ord_antisym.
    - apply join_lub;split.
      + apply bot_min;auto.
      + apply ord_refl.
    - apply join_ub_r.
  Qed.

  Lemma join_bot_r: ∀ a, a ⋁ ⊥ = a.
  Proof.
    intros.
    apply (ord_antisym).
    - apply @join_lub;split.
      + apply ord_refl.
      + apply bot_min;auto.
    - apply join_ub_l.
  Qed.


  Lemma join_top_l: ∀ a, ⊤ ⋁ a = ⊤.
  Proof.
    intros.
    apply ord_antisym.
    apply top_max.
    apply join_ub_l.
  Qed.
  
  Lemma join_top_r: ∀ a, a ⋁ ⊤ = ⊤.
  Proof.
    intros.
    apply ord_antisym.
    apply top_max.
    apply join_ub_r.
  Qed.

  Lemma meet_top_l: ∀ a, ⊤ ⋀ a = a.
  Proof.
    intros.
    apply ord_antisym.
    - apply meet_lb_r.
    - apply @meet_glb;split.
      + apply top_max.
      + apply ord_refl.
  Qed.

  Lemma meet_top_r: ∀ a, a ⋀ ⊤ = a.
  Proof.
    intros.
    apply ord_antisym.
    - apply meet_lb_l.
    - apply @meet_glb;split.
      + apply ord_refl.
      + apply top_max.
  Qed.

End BotTop.

Hint Resolve bot_min meet_bot_l meet_bot_r join_bot_l join_bot_r.
Hint Resolve top_max join_top_l join_top_r meet_top_l meet_top_r.

