Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import Morphisms.
Require Import ImpAlg.Ens.
Require Import ImpAlg.Lattices.
Require Import ImpAlg.BotTop.



Definition arrow_set_l `{X:Set} `{arrow:X → X → X} (A:Ens) (b:X):Ens :=
  fun x => ∃a, A a ∧ x = arrow a  b.
Definition arrow_set_r `{X:Set} `{arrow:X → X → X} (a:X) (B:Ens):Ens :=
    fun x => ∃b, B b ∧ x = arrow a b.
Definition arrow_set `{X:Set} `{arrow:X → X → X} (A:Ens) (B:Ens):Ens :=
    fun x => ∃a,∃b, A a ∧ B b ∧ x = arrow a b.

Global Instance arrow_set_Proper `{X:Set} `{arrow:X → X → X} a:
  Proper ((@same_set X) ==> (@same_set X)) ((@arrow_set_r X arrow) a).
Proof.
  intros A B HAB.
  split;intros Hx; destruct Hx as [b [Hb Hx]];
  exists b;
  split;[apply HAB,Hb| assumption|apply HAB,Hb| assumption].
Qed.


Class ImpStructure `{CL:CompleteLattice} :=
  {
    arrow:X → X → X;

    arrow_mon_l: ∀ a a' b, ord a a' ->  ord (arrow a' b ) (arrow a b );
    arrow_mon_r: ∀ a b b', ord b b' ->  ord (arrow a b  ) (arrow a b');

    arrow_meet: ∀ a B, arrow a (meet_set B) = meet_set (@arrow_set_r _ arrow a B)
  }.

Infix "↦" := arrow (at level 60, right associativity):ia_scope.
Notation "a ↪ B":= (@arrow_set_r _ _ a B) (at level 69):ia_scope.


Generalizable Variables X ord join meet arrow L meet_set join_set.

Section About_ImpStructure.
  Context `{IS:ImpStructure}.


  
  Lemma arrow_mon: ∀ a a' b b' , a' ≤ a -> b ≤ b' -> arrow a b ≤ arrow a' b'.
  Proof.
    intros.
    apply (ord_trans  _ (a' ↦ b)).
    - apply arrow_mon_l; auto.
    - apply arrow_mon_r; auto.
  Qed.
  
  Lemma arrow_prenex:
    ∀ a f, (λ y:X, ∃ b : X, y = a ↦ f a b) ≅ (@arrow_set_r X arrow a  (λ y : X, ∃ b : X, y = f a b)).
  Proof.
    intros.
    unfold arrow_set_l.
    split;intros;
    destruct H as [b Hb].
    - exists (f a b).
      split;[exists b | rewrite Hb];reflexivity.
    - destruct Hb as [Hb Hy].
      destruct Hb as [c Hc].
      exists c;rewrite Hy,Hc;reflexivity.
  Qed.


  Lemma meet_set_glb: ∀ A, (glb A) (⊓ A).
  Proof.
    intros A.
    split.
    red.
    apply meet_elim.
    apply meet_intro.
  Qed.

  Notation "∩ f" := (meet_family f) (at level 60).
  
  Lemma meet_fam_elim : ∀ I (f:I → X) (i:I), (∩ f)  ≤ f(i).
  Proof.
    intros.
    unfold meet_family.
    apply meet_elim.
    exists i.
    reflexivity.
  Qed.

  Lemma meet_fam_intro : ∀I f y, (∀ (i:I), y ≤ f i) → y ≤ ∩ f.
  Proof.
    intros.
    apply meet_intro.
    intros.
    destruct H0 as [i Hi];rewrite Hi; auto.
  Qed.


  Lemma meet_equality: ∀ A B, A ≅ B → ⊓ A = ⊓ B.
  Proof.
    intros.
    apply ord_antisym;
      apply meet_intro;intros x Hx;
      [destruct (H x) as [_ Ok]|destruct (H x) as [Ok _]];
      apply meet_elim; apply (Ok Hx).
  Qed.  


  Lemma double_meet :
    ∀ P, ⊓(λ x1 : X,∃ a0 : X, x1 = ⊓ (P a0))
         = ∩ (fun x => ⊓ (P x)).
  Proof.
    intros.
    reflexivity.
  Qed.


  Lemma double_meet_elim :
    ∀ (f:X → X → X) a b, ∩ (fun x => ⊓ (fun y => ∃ z, y= f x z)) ≤ f a b.
  Proof.
    intros f a b.
    apply (ord_trans _ ( ⊓ (λ y : X, ∃ z, y = f a z))).
    * apply (@meet_fam_elim _ (fun (x:X) => ⊓ (λ y : X, ∃ z : X, y = f x z))).
    * apply meet_elim.
      exists b;reflexivity.
  Qed.

End About_ImpStructure.











