Set Implicit Arguments.
Set Contextual Implicit.

Require Import Utf8.
Require Import Setoid.
Require Import Morphisms.
Require Import Relation_Definitions.
Require Import TLC.LibLN.
Require Import Arith.
Require Import ImpAlg.Lattices.
Require Import ImpAlg.ImpAlg.
Require Import ImpAlg.Ens.
Require Import ImpAlg.BotTop.
Require Import ImpAlg.ImpAlg_Lambda.
Require Import ImpAlg.Lambda.
Require Import ImpAlg.ImpAlg_Adequacy.


(*********************************)
(*       Separator axioms        *)
(*********************************)

Definition Succ := Datatypes.S.
Definition K `{IS:ImpStructure}:= ⋂[a] (⋂[b] (a ↦ b ↦ a)).
(*     meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = arrow a ( arrow b a))). *)
Definition S `{IS:ImpStructure}:= ⋂[a] (⋂[b] ⋂[c] ((a ↦ b ↦ c) ↦ (a ↦ b) ↦ a ↦ c)).
(*   meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = meet_set
      (fun z => ∃ c, z=(arrow (arrow a (arrow b c)) (arrow (arrow a b) (arrow a c)))))). *)
(** cc already defined in Adequacy       
 Definition cc `{IS:ImpStructure} :=⋂[a] (⋂[b] (((a ↦ b) ↦ a) ↦ a)). *)
(*   meet_set (fun x => ∃ a, x= meet_set (fun y => ∃ b, y = arrow (arrow (arrow a b) a) a)). *)



Class ImpAlg `{IS:ImpStructure}:=
  {
    separator: X → Prop;
    modus_ponens: ∀ a b, separator a → separator (arrow a b) → separator b;
    up_closed: ∀ a b, ord a b → separator a → separator b;
    sep_K : separator K;
    sep_S : separator S;
    sep_cc : separator cc
  }.

Generalizable Variables separator.

Notation "a ∈ ʆ":= (separator a) (at level 80,no associativity).
Infix "@" := (app) (at level 59, left associativity).
Notation "λ- f":= (abs f) (at level 60).


Hint Resolve arrow_meet modus_ponens sep_K sep_S sep_cc.
Hint Unfold arrow_set_l arrow_set_r arrow_set.




Ltac repeat_meet_intro :=
  repeat
    (apply meet_intro;intros ?x ?Hx;
     repeat (
         match goal with
           | Ha: exists x, _  |- _  => destruct Ha as (?x,?Hx)
           | Ha: ?a = _ |- _ ≤  ?a => rewrite Ha
           | Ha: _ ∧ _  |- _  => destruct Ha as (?Ha1,?Ha2)
      end)
     ).

Ltac auto_meet_elim :=
  apply meet_elim;
  match goal with
    | |- (exists !a, ?b  ↦ _ = ?a ↦ _ )=> exists ?b;reflexivity
    | |- _ => auto
  end.


Ltac find_translated :=

  repeat (
  match goal with
    | |- translated (trm_abs _ ) _ => 
      apply tr_abs;intros;unfold open_trm;simpl;unfold open_rec_trm;simpl;unfold Succ;repeat case_nat
    | |- translated (trm_app _ _) _ => apply tr_app
    | |- translated (trm_cvar _) _ => apply tr_var
  end
    ).



Ltac find_translated_typ :=
   repeat (
  match goal with
    | |- translated_typ (typ_fall _ ) _ =>
        apply tr_typ_fall;intros;unfold open_typ;unfold open_rec_typ
    | |- translated_typ (typ_arrow _ _) _ => apply tr_typ_arrow
    | |- translated_typ (typ_cvar _) _ => apply tr_typ_var
    | |- translated_typ (If ?a = ?a then _ else _ ) _ => rewrite~ If_l
    | |- translated_typ (_ _ (If Succ ?a = 0 then _ else _ ) _ ) _  => rewrite~ If_r
  end
     ).

 
Section Separator.
  Context `{IA:ImpAlg}.



  Definition env:=@env X.
  Definition codom:=@codom X.
  Definition fv_trm:=@fv_trm X.
  Definition fv_typ:=@fv_typ X.

  
  
  Ltac adjunction_mon:=
  repeat
    match goal with
      | |- (_ ↦ ?a ≤ _ ↦ ?a) => apply arrow_mon_l
      | |- (?a ↦ _ ≤ ?a ↦ _ )=> apply arrow_mon_r
      | |- _@?a ≤ _@?a       => apply app_mon_l
      | |- _  ≤ ?a ↦ (_@?a)=> apply adjunction
      | |- ?a ≤ _ ↦ (?a@_)=> apply adjunction
      | |- (?a ↦ _)@?a ≤ _ => apply adjunction
      | |- (_ ↦ ?a)@_ ≤ ?a => apply adjunction
    end;try (apply ord_refl); auto.

  Ltac gather_vars :=
    let A := gather_vars_with (fun x : vars => x) in
    let B := gather_vars_with (fun x : var => \{x}) in
    let C := gather_vars_with (fun x : env => (dom x)\u (codom x)) in
    let D := gather_vars_with (fun x : trm => fv_trm x) in
    let E := gather_vars_with (fun x : typ => fv_typ x) in
    constr:(A \u B \u C \u D \u E).
  
  (** Tactic [pick_fresh x] adds to the context a new variable x
    and a proof that it is fresh from all of the other variables
    gathered by tactic [gather_vars]. *)
  
  Ltac pick_fresh Y :=
    let L := gather_vars in (pick_fresh_gen L Y).
  
  (** Tactic [apply_fresh T as y] takes a lemma T of the form 
    [forall L ..., (forall x, x \notin L, P x) -> ... -> Q.]
    instanciate L to be the set of variables occuring in the
    context (by [gather_vars]), then introduces for the premise
    with the cofinite quantification the name x as "y" (the second
    parameter of the tactic), and the proof that x is not in L. *)

  Tactic Notation "apply_fresh" constr(T) "as" ident(x) :=
    apply_fresh_base T gather_vars x.
  Tactic Notation "apply_fresh" "*" constr(T) "as" ident(x) :=
    apply_fresh T as x; auto_star.
  Tactic Notation "apply_fresh" constr(T) :=
    apply_fresh_base T gather_vars ltac_no_arg.
  Tactic Notation "apply_fresh" "*" constr(T) :=
    apply_fresh T; auto_star.


  Ltac fresh_term_abs :=
    let L0:=gather_vars in let L1 := beautify_fset L0 in
                           apply (@term_abs _ L1);intros.
  Ltac fresh_typing_abs :=
    let L0:=gather_vars in let L1 := beautify_fset L0 in
                           apply (@typing_abs _ L1);intros.
  Ltac autoterm :=
  repeat (
      try (apply term_cc);
      try (apply term_var);
      try (fresh_term_abs);
      try (apply term_app);
      try (unfold Succ);
      try (case_nat)
    ).

 
 
 
 
 
  (******************************************)
  (*          Combinatory Logic             *)
  (******************************************)
  
  Notation "λ+ t":= (trm_abs (t:trm)) (at level 69, right associativity).
  Notation "[ t ] u":= (trm_app (t:trm) u ) (at level 69, right associativity).
  Notation "$ n":= (@trm_bvar X n) (at level 69, right associativity).
  
 
  Definition Id:= ⊓ (fun y => ∃ a, y = a ↦ a).
  Definition lambdaId:=  (λ+ $0).

  Lemma lambda_Id: Id = λ- (fun x => x).
  Proof.
    unfold abs.
    reflexivity.
  Qed.


  Lemma translated_Id : translated lambdaId Id.
  Proof.
    unfold lambdaId.
    find_translated.
  Qed.  
  

  (* Proving that [λxy.x] = ⊓ ⊓ (a ↦ b ↦ a) *)
  
  (* Definition K:= ⊓ (fun x => ∃ a, x= ⊓ (fun y => ∃ b, y = a ↦ b ↦ a )). *)
  
  Definition lambdaK:= (λ+ λ+ $1).
  

  Lemma lambda_K:
    K = λ- (fun x=> λ- (fun y => x)).
  Proof.
    unfold abs.
    apply meet_equality.
    split;intros Hx;
    destruct Hx as [a Ha];
    exists a;
    rewrite Ha,arrow_meet in *;
    apply meet_equality;
    [ | apply same_set_sym];
    apply arrow_prenex with (f:=fun x => (fun y => y ↦ x)).
  Qed.

  Lemma translated_K : translated lambdaK K.
  Proof.
    rewrite lambda_K.
    unfold lambdaK.
    find_translated.
  Qed. 


  (* Proving that [λxyz.x y (z y)] = ⊓ ⊓ ((a ↦ b ↦ c) ↦ (a ↦ b) ↦ a ↦ c) *)

  (* Definition S:= *)
  (*   ⊓(fun x => ∃ a, x = ⊓(fun y => ∃ b, y = ⊓(fun z => ∃ c , z = (a ↦ b ↦ c) ↦ (a ↦ b) ↦ a ↦ c))). *)
  Definition lambdaS:=
    λ+ λ+ λ+ [[$2] $0] ([$1] $0).


  Lemma lambda_S: S = λ- (fun x => λ- (fun y => λ- (fun z => (x@z)@(y@z)))).
  Proof.
    unfold S.
    apply (ord_antisym).
    - apply meet_intro;intros x [a Ha].
      unfold abs in Ha at 1.
      rewrite arrow_meet in Ha.
      rewrite Ha.
      apply meet_intro;intros y [z [[b Hb] Hy]].
      rewrite Hy,Hb.
      unfold abs.
      do 2 (rewrite arrow_meet).
      apply meet_intro;intros w [v [[u [[c Hc] Hv]] Hw]].
      rewrite Hw,Hv,Hc.
      clear Ha Hb Hy Hc Hv Hw w v u y z x.

      apply (ord_trans _ (⊓ (λ z : X, ∃ d : X, z = (c ↦ (b@c) ↦ d) ↦ (c ↦ (b@c)) ↦ c ↦ d))).
      + apply (double_meet_elim (fun c => fun b => ⊓ (λ z : X, ∃ d : X, z = (c ↦ b ↦ d) ↦ (c ↦ (b)) ↦ c ↦ d))).
      + unfold app at 3.
        repeat (rewrite arrow_meet).
        apply meet_intro;intros z [y [[w [[d [Hd Hw]] Hy]] Hz]].
        rewrite Hz,Hy,Hw.
        apply (ord_trans _ ((c ↦ (b@c) ↦ d) ↦ (c ↦ (b@c)) ↦ c ↦ d)).
        * apply meet_elim.
          exists d; reflexivity.
        * apply arrow_mon.
          apply adjunction.
          apply Hd.
          adjunction_mon.
          
    - repeat_meet_intro.
      unfold abs.
      apply (ord_trans _ ((a ↦ b ↦ c) ↦ ⊓(λ x1 : X,∃ a1 : X,
                                                          x1 = a1 ↦ ⊓ (λ x2 : X, ∃ a2 : X, x2 = a2 ↦ (a ↦ b ↦ c) @ a2 @ (a1 @ a2))))).
      + apply meet_elim.
        exists (a ↦ b ↦ c).
        reflexivity.
      + apply arrow_mon_r.
        apply (ord_trans _ ((a ↦ b) ↦  ⊓ (λ x2 : X, ∃ a2 : X, x2 = a2 ↦ (a ↦ b ↦ c) @ a2 @ ((a ↦ b) @ a2)))).
        * apply meet_elim.
          exists (a ↦ b).
          reflexivity.
        * apply arrow_mon_r.
          apply (ord_trans _ (a ↦ (a ↦ b ↦ c) @ a @ ((a ↦ b) @ a))).
          apply meet_elim;exists a;reflexivity.
          repeat (try (apply arrow_mon_r);try (apply arrow_mon_l);apply adjunction).
          apply ord_refl.
  Qed.

  Lemma translated_S: translated lambdaS S.
  Proof.
    rewrite lambda_S.
    unfold lambdaS.
    find_translated.
  Qed.


  (* Proving that I = S K K *)

  Lemma Id_SK : (S @ K) @ K ≤ Id.
  Proof.
    rewrite lambda_S.
    apply (ord_trans _ ((λ- (λ y : X, λ- (λ z : X, K @ z @ (y @ z)))) @ K )).
    - apply app_mon_l.
      apply (betarule (fun x => (λ- (λ y : X, λ- (λ z : X, x @ z @ (y @ z))))) K). 
    - apply (ord_trans _ (λ- (λ z : X, K  @ z @ (K  @ z)))).
      apply (betarule  (fun y =>  λ- (λ z : X, K  @ z @ (y @ z))) K). 
      rewrite lambda_Id.
      repeat_meet_intro.
      apply adjunction.
      apply (ord_trans _ (K @ a @ (K @ a))).
      apply (betarule (fun z => K @ z @ (K @ z)) a).
      unfold K at 1.
      do 2 (apply adjunction).
      apply (double_meet_elim (fun x => fun y => x ↦ y ↦ x)).
  Qed.



  (* Proving that [λxy.xy] = ⊓ ⊓ ((a ↦ b) ↦ a ↦ b)) *)
  (* In fact, Easy := S@(S@(K@S)@(S@(K@K)@I))@(K@I)  *)
  Definition Easy:= ⊓ (λ x:X, ∃ a, x= ⊓ (λ y:X, ∃ b, y = ( (a ↦ b) ↦ a ↦ b))).
  Lemma lambda_Easy:
    Easy = λ- (fun x => λ- (fun y => x @ y)).
  Proof.
    unfold abs,Easy.
    apply ord_antisym.
    -  apply meet_intro;intros.
       destruct H as [a Ha].
       rewrite arrow_meet in Ha.
       rewrite Ha.
       apply meet_intro;intros.
       red in H.
       destruct H as [b [[c Hb] Hx0]].
       rewrite Hx0,Hb.
       unfold app.
       do 2 (rewrite arrow_meet).     
       apply meet_intro; intros w [y [[z [Hz Hy]] Hw]].
       rewrite Hw,Hy.
       rewrite (double_meet ).
       apply (ord_trans _ ((c ↦ z) ↦ c ↦ z)).
       + apply (double_meet_elim (fun x => fun y => (x ↦ y) ↦ x ↦ y)).
       + apply arrow_mon_l.
       apply Hz.

    - apply meet_intro;intros.
      destruct H as [a Ha].
      rewrite Ha.
      apply meet_intro;intros b [c Hb].
      rewrite Hb.
      apply (ord_trans _ ((a ↦ c) ↦ ⊓ (λ x0 : X, ∃ a0 : X, x0 = a0 ↦ (a ↦ c) @ a0))).
      + apply meet_elim.
        exists (a ↦ c).
        reflexivity.
      + apply arrow_mon_r.
        apply (ord_trans  _ (a ↦ (a ↦ c) @ a)).
        * apply meet_elim.
          exists a;reflexivity.
        * apply arrow_mon_r.
          apply meet_elim.
          apply ord_refl.
  Qed.

  Definition lambdaEasy:= λ+ λ+ [$1] $0.
  Lemma translated_Easy: translated lambdaEasy Easy.
  Proof.
    unfold lambdaEasy.
    rewrite lambda_Easy.
    find_translated.
  Qed.  

  (* It is not the real W... *)
  Definition W := ⊓ (λ x:X, ∃ a, x= ⊓ (λ y:X, ∃ b, y = (a ↦ a ↦ b) ↦ a ↦ b)).
  Lemma lambda_W:
    W = λ- (fun x => λ- (fun y => (x @ y) @ y)).
  Proof.
    unfold abs,W.
    apply ord_antisym.
    - repeat_meet_intro.
      rewrite arrow_meet. 
      apply meet_intro.
      intros y [z [[c Hz] Hy]].
      rewrite Hy,Hz.
      clear y z Hy Hz x Hx.
      unfold app.
      do 2 (rewrite arrow_meet).     
      apply meet_intro; intros w [y [[z [Hz Hy]] Hw]].
      rewrite Hw,Hy.
      rewrite (double_meet).
      apply (ord_trans _ ((c ↦ c ↦ z) ↦ c ↦ z)).
      + apply (double_meet_elim (fun x => fun y => (x ↦ x ↦ y) ↦ x ↦ y)).
      + apply arrow_mon_l.
        apply adjunction.
        apply Hz.
    - repeat_meet_intro.
      apply (ord_trans _ ((a ↦ a ↦ b) ↦ ⊓ (λ x0 : X, ∃ a0 : X, x0 = a0 ↦ (a ↦ a ↦ b)@ a0 @ a0))).
      + apply meet_elim.
        exists (a ↦ a ↦ b).
        reflexivity.
      + apply arrow_mon_r.
        apply (ord_trans  _ (a ↦ (a ↦ a ↦ b) @ a @ a)).
        * apply meet_elim.
          exists a;reflexivity.
        * adjunction_mon.
          apply adjunction.
          adjunction_mon.
  Qed.

  Definition lambdaW:= λ+ λ+ [[$1] $0] $0.
  Lemma term_W: term lambdaW.
  Proof.
      fresh_term_abs.
      try (apply term_cc);
      try (apply term_var);

      try (apply term_app);
      try (unfold Succ);
      try (case_nat).

    autoterm.
  Qed.
  
  Lemma translated_W: translated lambdaW W.
  Proof.
    unfold lambdaW.
    rewrite lambda_W.
    find_translated.
  Qed.


  Hint Resolve lambda_S lambda_K lambda_Id translated_Id translated_S translated_K.


  (***********************************************************)


  Inductive combinator : X → Prop :=
  | comb_S : combinator S
  | comb_K : combinator K
  | comb_Id : combinator Id
  | comb_app: ∀ A B, combinator A → combinator B → combinator (A@B).                      

  Inductive termSK : trm → Prop :=
  |term_S : termSK lambdaS
  |term_K : termSK lambdaK
  |termSK_app: ∀ t u, termSK t → termSK u → termSK (trm_app t u).

  Hint Resolve comb_S comb_K comb_Id comb_app term_S term_K termSK_app.

  Lemma combinator_translated: ∀ t C, termSK t → translated t C → combinator C.
  Proof.
    intros t C SK.
    generalize C.
    clear C.
    induction SK;intros C Tr. 
    - replace C with S.
      auto.
      apply (@translated_unique X _ _ _ _ lambdaS _ C);auto.
    - replace C with K.
      auto.
      apply (@translated_unique _ _ _ _ _ lambdaK _ C);auto.
    - inversion Tr;auto.
  Qed.

  Lemma combinatorSK: ∀ t, termSK t → ∃ u, combinator u ∧ translated t u.
  Proof.
    intros t SK.
    induction SK.
    - exists S;auto.
    - exists K;auto.
    - destruct IHSK1 as [Ct [HCt HTt]].
      destruct IHSK2 as [Cu [HCu HTu]].            
      exists (Ct@Cu).
      split;auto.
      apply tr_app;auto.
  Qed.



  Definition neg a:=a ↦ ⊥.
  Definition closed_typ (T:typ):=  fv_typ T = \{}.


  (**************************************************)
  (*                                                *)
  (*  Bunch of tactics to automatize typing, etc..  *)
  (*                                                *)
  (**************************************************)

  
  Ltac autofv := cbn;repeat (try rewrite union_empty_l;try rewrite union_empty_r);reflexivity.
  Ltac autoclosed:=split;[autoterm|autofv].
  Ltac translated_closed:=split;[find_translated|autoclosed].
  Ltac exists_term t:= exists t;translated_closed.
  Ltac autoclosed_typ:=unfold closed_typ;cbn;repeat (rewrite union_empty_l);auto.

  


  Ltac typ_from_translated:=
    unfold neg;
    repeat (
        match goal with
          | |- translated_typ  _ (⊓ (_)) => apply tr_typ_fall;intros
          | |- translated_typ _ (_ ↦ _) => apply tr_typ_arrow
          | |- translated_typ _ ?a => apply tr_typ_var
        end
      ).

  
  Ltac autotype :=
    repeat (
        try (apply type_fvar);
        try (apply type_cvar);
        try (apply_fresh type_fall);
        try (apply type_arrow)
      ).

  Ltac simplIf:=repeat (try (rewrite If_l;[|reflexivity]);try (rewrite If_r;[|auto])).
  
  Ltac autoget:=
    repeat(
        match goal with
          | |- _ # _ =>  apply get_none_inv
          | |- get _ empty = None => apply get_empty
          | |- get _ (_ & _ ) = None => rewrite get_push;simplIf
        end
      ).
  Ltac autook:=repeat (apply ok_push);try (apply ok_empty); autoget.

  Ltac autobinds:=
    match goal with
      | |- binds ?y _ (?E & ?y ~ _) => apply* binds_push_eq
      | |- binds ?y _ (?E & ?z ~ _) => apply* binds_push_neq
    end.

  Ltac autotyping:=
    repeat (
      match goal with
        | H: _ \notin _ \u _ |- _ => rewrite_all notin_union in * 
        | H: _ \notin \{_}   |- _ => rewrite_all notin_singleton in *
        | H: _ \notin \{}    |- _ => clear H
        | H: _ ∧ _           |- _ => destruct H
        | |- ok _ =>  autook
        | |- typing_trm _ (λ+ _)  (typ_arrow _ _) => fresh_typing_abs;autotype;unfold open_trm;simpl;simplIf
        | |- typing_trm _ ([ _ ] _ ) _ => eapply typing_app 
        | |- typing_trm _ (trm_fvar _) ?a => apply typing_prf_var;[autook|autobinds|autotype]
        | |- type _ => autotype
        end
      ).
 
  

  Definition closed (t:trm):= term t ∧ fv_trm t = \{}.

  Lemma typing_empty_closed: ∀ t T, typing_trm empty t T → closed t.
  Proof.
    intros t T H;split.
    apply (typing_term H).
    apply (typing_fv_trm H).
  Qed.

  (** This is our only axiom, to avoid the definition of the compilation functions
      from λ-terms to combinator *)

  Axiom lambdaSK : ∀ t, closed t → termSK t.

  Lemma combinatorLambda: ∀ t, closed t → ∃ u, combinator u ∧ translated t u.
  Proof.
    intros.
    apply combinatorSK.
    apply lambdaSK.
    assumption.
  Qed.

  Lemma combinator_trm_translated: ∀ t C, closed t → translated t C → combinator C.
  Proof.
    intros t C SK Tr.
    apply (@combinator_translated t);auto.
    apply lambdaSK;auto.
  Qed.



  (*******************************)
  (*    Separator properties     *)
  (*******************************)

  Definition entails a b := (a ↦ b) ∈ ʆ .
  Infix "⊢":= entails (at level 75).
  Definition equivalence a b := a ⊢ b ∧  b ⊢ a.
  Infix "≈":= equivalence (at level 80).

  Lemma equiv_sep: ∀ a b, a ≈ b → (a ∈ ʆ ↔ b ∈ ʆ).
  Proof.
    intros a b (Hab,Hba).
    split;intro H;
    [apply (modus_ponens _ _ H Hab)|
     apply (modus_ponens _ _ H Hba)].
  Qed.

  Lemma app_closed : ∀ a b, a ∈ ʆ → b ∈ ʆ → a@b ∈ ʆ.
  Proof.
    intros.
    apply (modus_ponens b (a@b)).
    - assumption.
    - apply (up_closed a).
      + apply adjunction.
        apply ord_refl.
      + assumption.
  Qed.


  Hint Resolve sep_S sep_K.
  Lemma sep_Id: Id ∈ ʆ.
  Proof.
    apply (up_closed  _ _ Id_SK).
    repeat (apply app_closed);auto.
  Qed.

  Hint Resolve sep_Id.
  
  Lemma sep_comb: ∀ A, combinator A → A ∈ ʆ.
  Proof.
    intros.
    induction H;auto.
    apply app_closed;auto.
  Qed.

  Lemma sep_SK: ∀ C, (exists t, translated  t C ∧ termSK t) → C ∈ ʆ.
  Proof.
    intros.
    apply sep_comb.
    destruct H as [t [HC Ht]].
    apply (combinator_translated Ht HC).
  Qed.
  
  Lemma sep_term: ∀ C, (exists t, translated  t C ∧ closed t) → C ∈ ʆ.
  Proof.
    intros.
    apply sep_comb.
    destruct H as [t [HC Ht]].
    apply (combinator_trm_translated Ht HC).
  Qed.
  Ltac in_sep_term t:= apply sep_term;exists t;split;[find_translated|autoclosed].


  Lemma sep_typ: ∀ C, (exists t T, translated_typ T C ∧ closed_typ T ∧ typing_trm empty t T) → C ∈ ʆ.
  Proof.
    intros C (t,(T,(Tr,(Cl,Typ)))).
    destruct (combinatorLambda (typing_empty_closed Typ) ) as (u,(Comb,Htu)).
    eapply up_closed.
    apply (adequacy_empty Typ Htu Cl Tr).
    now apply sep_comb.
  Qed.


  Lemma sep_W: W ∈ ʆ.
  Proof.
    apply sep_term.
    exists lambdaW.
    split.
    apply translated_W.
    autoclosed.
  Qed.

  Lemma sep_Easy: Easy ∈ ʆ.
  Proof.
    apply sep_term.
    exists lambdaEasy.
    split.
    apply translated_Easy.
    autoclosed.
  Qed.

  Hint Resolve sep_Id sep_W.

  Lemma entails_refl : ∀ a, a ⊢ a.
  Proof.
    unfold entails.
    intros.
    apply (up_closed (Id)).
    - apply meet_elim.
      exists a;reflexivity.
    - auto.
  Qed.

  Lemma subtyping : ∀ a b, a ≤ b → a ⊢ b.
  Proof.
    intros.
    apply (up_closed (a ↦ a)).
    + apply arrow_mon_r.
      assumption.
    + apply entails_refl.
  Qed.


  Lemma meet_impredicative: ∀ a b c, a ↦ b ↦ c ⊢ (a ⋀ b) ↦ c.
  Proof.
    intros.
    apply (up_closed W);auto.
    unfold W.
    apply (ord_trans _ (((a ⋀ b) ↦ (a ⋀ b) ↦ c) ↦ (a ⋀ b) ↦ c)).
    - rewrite (double_meet).
      apply (double_meet_elim (fun x => fun y => ((x ↦ x ↦ y) ↦ x ↦ y))).
    - apply arrow_mon_l.
      apply arrow_mon.
      apply meet_lb_l.
      apply arrow_mon_l.
      apply meet_lb_r.
  Qed. 

  
  
  Lemma mod_ponens : ∀ a b, (a ↦ b) ⋀ a ⊢ b.
  Proof.
    intros.
    apply (modus_ponens ((a ↦ b) ↦  a  ↦ b)).
    - apply (up_closed Easy).
      unfold Easy.
      rewrite (double_meet).
      apply (double_meet_elim (fun x => fun y => (x ↦ y) ↦ x ↦ y)).
      apply sep_Easy.
    -apply meet_impredicative.
  Qed.


  Lemma exfalso_S: ∀ a, ⊥ ⊢ a.
  Proof.
    intro.
    apply (subtyping _ _ (bot_min a)).
  Qed.

  Lemma true_S: ∀ a, a ⊢ ⊤.
  Proof.
    intro.
    apply (subtyping _ _ (top_max a)).
  Qed.

  Lemma realizer_later: ∀ a, (exists b, b ∈ ʆ ∧ b ≤ a) → a ∈ ʆ.
  Proof.
    intros a (b,(Hb,Ha)).
    apply (modus_ponens b _ Hb).
    apply* (subtyping).
  Qed.

  Ltac emeet_elim:=apply (meet_elim_trans);later;split;[later;reflexivity|].
  Ltac hnform:=repeat (emeet_elim;apply arrow_mon_r);adjunction_mon.
  Ltac realizer t:= apply realizer_later;later;split;[in_sep_term t| hnform].


  Lemma entails_trans : transitive X entails.
  Proof.
    unfold transitive.
    unfold entails;intros a b c Hab Hbc.
    apply (modus_ponens (b ↦ c));auto.
    apply (modus_ponens (a ↦ b));auto.
    realizer (λ+ λ+ λ+ ([$1] ([$2] $0))).
  Qed.


  Lemma modus_ponens_later: ∀ a, (exists b, b ∈ ʆ ∧ b ⊢ a ) → a ∈ ʆ.
  Proof.
    intros a (b,(Hb,Ha)).
    apply* (modus_ponens b).
  Qed.

  Notation "¬ a":=(neg a).

  Lemma dni: ∀ a, a ⊢ ¬ (¬ a).
  Proof.
    intro.
    unfold neg.
    realizer (λ+ λ+ ([$0] $1)).
  Qed.

  Hint Resolve (type_cvar).
  Lemma dne: ∀ a, (¬ (¬ a)) ⊢ a.
  Proof.
    intro.
    unfold entails,neg.
    apply (up_closed (((a ↦ ⊥ ) ↦ a) ↦ a)).
    - adjunction_mon.
    - apply sep_typ.    
      exists (@trm_cc X).
      later.
      repeat split.
      + do 3 (apply tr_typ_arrow;[|apply tr_typ_var]).
        apply tr_typ_var.
      + autoclosed_typ.
      + apply typing_cc;auto.
  Qed.


  Lemma dn: ∀ a, a ≈ ¬ ¬ a.
  Proof.
    intros;split;[apply dni|apply dne].
  Qed.


 
  Lemma contraposition_l: ∀ a b, a ↦ b ⊢ (¬ b) ↦ ¬ a.
  Proof.
    intros.
    unfold entails,neg.
   (* This is only to illustrate some tactics use, 
actually the following is enough 
    realizer ((λ+ λ+ λ+ ([ $ 1 ] ([$ 2] $ 0)))).*)
    apply sep_typ.
    exists ((λ+ λ+ λ+ ([ $ 1 ] ([$ 2] $ 0)))).
    later.
    repeat split.
    + typ_from_translated.
    + autoclosed_typ.
    + autotyping.
  Qed.




  Lemma arrow_entails_mon_r: ∀ a b b', b ⊢ b' → a ↦ b ⊢ a ↦ b'.
  Proof.
    intros a b b' H.
    apply modus_ponens_later.
    later.
    split.
    - eapply (app_closed _ (b ↦ b'));[|exact H].
      realizer (λ+ λ+ λ+ ([ $2 ] ([ $1 ] $0))).
    - apply subtyping.
      apply adjunction.
      repeat (apply arrow_mon_r).
      adjunction_mon.
  Qed.

  Lemma arrow_entails_mon_l: ∀ a a' b, a ⊢ a' → a' ↦ b ⊢ a ↦ b.
  Proof.
    intros a a' b H.
    apply modus_ponens_later.
    later.
    split.
    - eapply (app_closed _ (a ↦ a'));[|exact H].
      realizer (λ+ λ+ λ+ ([ $1 ] ([ $2 ] $0))).
    - apply subtyping.
      apply adjunction.
      repeat (apply arrow_mon_r).
      adjunction_mon.
  Qed.

  Lemma arrow_entails_mon: ∀ a a' b b', a' ⊢ a → b ⊢ b' → a ↦ b ⊢ a' ↦ b'.
  Proof.
    intros a a' b b' Ha Hb.
    apply modus_ponens_later.
    later.
    split.
    - eapply (app_closed _ (b ↦ b'));[|exact Hb].
      eapply (app_closed _ (a' ↦ a));[|exact Ha].
      apply sep_term.
      exists_term (λ+ λ+ λ+ λ+ ([ $2 ] ([ $1 ] ([ $3 ] $0)))).
    - apply subtyping.
      do 2 (apply adjunction).
      hnform.
  Qed.

  Lemma contraposition_r: ∀ a b, (¬ a) ↦ (¬ b) ⊢ b ↦ a.
  Proof.
    intros.
    apply (@entails_trans ((¬ a) ↦ ¬ b) ((¬ ¬ b) ↦ ¬ ¬ a)).
    - apply contraposition_l.
    - apply arrow_entails_mon.
      + apply dni.
      + apply dne.
  Qed.

End Separator.